import json
import os
from itertools import islice
import numpy as np
import struct
import warnings
import re

from nextnanopy.utils.datasets import Variable, Coord
from nextnanopy.utils.mycollections import DictList
from nextnanopy.utils.formatting import best_str_to_name_unit
from nextnanopy.utils.misc import get_filename, message_decorator, start_with_choice
from nextnanopy import defaults

import pyvista as pv


_msgs = defaults.messages["load_output"]
load_message = lambda method: message_decorator(
    method, init_msg=_msgs[0], end_msg=_msgs[1]
)


def displayname(data):
    """
    Parameters
    ----------
    data: DataFolder or filepath
    Returns
    -------
    str
    formatting filenames and folder names in datafolder to display DataFolder in tree structure
    """
    if isinstance(data, DataFolder):
        return os.path.basename(data.fullpath) + r"/"
    else:
        return os.path.basename(data)


class DataFolder(object):
    """
    This class stores information about output directory.
    The stored data contains files (.files) and folders (.folders)
    Navigation between folders could be done in 3 ways:
        1. DataFolder.folders['folder_name']
        2. DataFolder.go_to('subfolder1', 'subfolder2', 'subfolder3')
        3. DataFolder.subfolder1.subfolder2.subfolder3
    For each method see details below.

    The initialization of the class will execute load and create_navigation methods.


    Parameters:
    ----------------
    fullpath: str
        path to a file



    Attributes:
    ----------------
    fullpath: str
    path to a file

    folders: DictList
        subfolders of the DataFolder,
        keys: str
            names of subfolders
        values: DataFolder
            DataFolder objects of subfolders

    files: list
        paths to files in folder


    Methods:
    --------------
    load():
        load a DataFolder

    create_navigation:
        creates attributes for navigation like DataFolder.subfolder1.subfolder2.subfolder3
        if name of subfolder corresponds to existed attribute of DataFolder class, attribute will not be created!
        if name of subfolder constis spaces, dots or specials charecters, attribute will be created, but navigation
        to whis subfolder will not work  - attribute error.

    find(template, deep = False):
        searches for a files which names contain template.
        template shoud be string.
        if deep = True, searches in subfolders as well.

        return: list of files

    go_to(*args):
        goes to the location
        DataFolder_path\\arg1\\arg2\\arg3...

        if location is a file:
            return filepath
        if location is a folder:
            return DataFolder(location)

    filenames()
        return filenames of files in folder
    """

    def __init__(self, fullpath):
        if not os.path.isdir(fullpath):
            raise ValueError(f"{fullpath} is not a directory")
        self.fullpath = fullpath
        self.files = []
        self.folders = DictList()
        self.load()
        self.create_navigation()

    def load(self):
        list_of_nodes = sorted(os.listdir(self.fullpath))
        for node in list_of_nodes:
            node_path = os.path.join(self.fullpath, node)
            if os.path.isdir(node_path):
                new_folder = DataFolder(node_path)
                self.folders[node] = new_folder
            else:
                self.files.append(node_path)

    def create_navigation(self):
        check_list = dir(self)
        for key, folder in self.folders.items():
            if key in check_list:
                warnings.warn(
                    f"foldername '{key}' is not availabel for attribute navigation."
                    " Please, use DataFolder.go_to()"
                )
            else:
                setattr(self, key, folder)

    def find(self, template, deep=False):
        list_of_files = [
            file for file in self.files if template in os.path.basename(file)
        ]
        if not deep:
            return list_of_files
        if not self.folders:
            return list_of_files
        else:
            for folder in self.folders:
                list_of_files += folder.find(template=template, deep=deep)
            return list_of_files

    def find_multiple(self, templates, deep=False):
        list_of_files = []

        for file in self.files:
            if all(keyword in os.path.basename(file) for keyword in templates):
                list_of_files.append(file)

        if deep and self.folders:
            for folder in self.folders:
                list_of_files += folder.find_multiple(templates=templates, deep=deep)

        return list_of_files

    def file(self, filename):
        matched_files = self.find(template=filename, deep=False)
        if not matched_files:
            raise ValueError(
                f"No file with filename {filename} in directory {self.fullpath}"
            )
        elif len(matched_files) == 1:
            return matched_files[0]
        else:
            warnings.warn(
                f"More than one file match '{filename}' name match in directory {self.fullpath}. First match returned."
            )
            return matched_files[0]

    def go_to(self, *args):
        path = os.path.join(self.fullpath, *args)
        if os.path.isdir(path):
            data = DataFolder(path)
        elif os.path.isfile(path):
            data = path
        else:
            raise ValueError(f"No such file or directory {path}")
        return data

    def filenames(self):
        return [os.path.basename(file) for file in self.files]

    def __repr__(self):
        out = list()
        out.append(f"{self.__class__.__name__}")
        out.append(f"fullpath: {self.fullpath}")
        out.append(f"Folders: {len(self.folders)}")
        for key in self.folders.keys():
            out.append(key)
        out.append("Files:")
        out.append(str(self.filenames()))
        out = "\n".join(out)
        return out

    def make_tree(self, level=0, result=None, with_files=True, deep=True):
        if not result:
            result = []
        result.insert(level, level * "    " + displayname(self))
        if with_files:
            for file in reversed(self.files):
                result.insert(level + 1, (level + 1) * "    " + displayname(file))
        level = level + 1
        for i in reversed(range(len(self.folders))):
            folder = self.folders[i]
            if deep:
                result = folder.make_tree(level, result=result, with_files=with_files)
            else:
                result.insert(level, level * "    " + displayname(folder))
        return result

    def show_tree(self, with_files=True, deep=True):
        tree_list = self.make_tree(with_files=with_files, deep=deep)
        print("\n".join(tree_list))

    @property
    def name(self):
        # folder_path = os.path.dirname(self.fullpath)
        folder_name = os.path.basename(self.fullpath)
        return folder_name

    def read_sweep_infodict(self):
        infodict_path = os.path.join(self.fullpath, "sweep_infodict.json")

        if not os.path.isfile(infodict_path):
            raise FileNotFoundError(
                "Sweep infodict file not found. Make sure that the chosen folder is a sweep output folder."
            )

        with open(infodict_path, "r") as file:
            infodict = json.load(file)

        return infodict


class Output(object):

    def __init__(self, fullpath, **loader_kwargs):
        self.fullpath = fullpath
        self.metadata = {}
        self.coords = DictList()
        self.variables = DictList()

    @property
    def folder(self):
        return os.path.split(self.fullpath)[0]

    @property
    def filename_only(self):
        return get_filename(self.fullpath, ext=False)

    @property
    def filename(self):
        return get_filename(self.fullpath, ext=True)

    @property
    def extension(self):
        return os.path.splitext(self.fullpath)[-1]

    @property
    def data(self):
        dl = DictList()
        dl.update(self.coords)
        dl.update(self.variables)
        return dl

    @load_message
    def load(self):
        pass

    def get_coord(self, key):
        return self.coords[key]

    def get_variable(self, key):
        return self.variables[key]

    def __getitem__(self, item):
        return self.data[item]

    def __setitem__(self, item, value):
        if item in self.coords.keys():
            self.coords[item] = value
        if item in self.variables.keys():
            self.variables[item] = value

    def __delitem__(self, item):
        if item in self.coords.keys():
            del self.coords[item]
        if item in self.variables.keys():
            del self.variables[item]

    def __repr__(self):
        out = []
        out.append(f"{self.__class__.__name__}")
        out.append(f"fullpath: {self.fullpath}")
        out.append(f"Coordinates: {len(self.coords)} datasets")
        for key, coord in self.coords.items():
            out.append(f"\t{str(coord)}")
        out.append(f"Variables: {len(self.variables)} datasets")
        for key, var in self.variables.items():
            out.append(f"\t{str(var)}")
        out = "\n".join(out)
        return out

    def __iter__(self):
        self._iter_index = 0
        return self

    def __next__(self):
        try:
            result = self.data.__getitem__(self._iter_index)
        except (IndexError, KeyError):
            raise StopIteration
        self._iter_index += 1
        return result


class DataFileTemplate(Output):
    """
    This class stores the data from any kind of nextnano data files with the
    same structure.
    The stored data contains coordinates (.coords) and dependent variables (.variables).
    Each coordinate or variable would contain attributes like name, unit and value.
    For more information, see their specific documentation.

    The initialization of the class will execute the load method.

    ...

    Parameters
    ----------
    fullpath : str
        path to the file.


    Attributes
    ----------
    fullpath : str
        path to the file (default: None)
    coords : DictList
        Coord objects (default: DictList())
    variables : DictList
        Variable objects (default: DictList())
    data : DictList
        coords and variables together
    metadata : dict
        extra information
    filename : str
        name with the file extension
    filename_only :
        name without the file extension
    extension : str
        file extension
    folder : str
        folder of the fullpath
    product : str
        flag about nextnano product to help to find the best loading routine


    Methods
    -------
    load(fullpath)
        load a data file

    get_coord(name)
        equivalent to self.coords[name]

    get_variable(name)
        equivalent to self.variables[name]

    """

    def __init__(self, fullpath, product=None, **loader_kwargs):
        super().__init__(fullpath)
        self.product = product
        self.load(**loader_kwargs)

    @load_message
    def load(self, **loader_kwargs):
        """
        Find the loader and update the stored information with the loaded data
        """
        loader = self.get_loader()
        df = loader(self.fullpath, **loader_kwargs)
        self.update_with_datafile(df)
        del df

    def update_with_datafile(self, datafile):
        """
        Copy .metadata, .coords and .variables of the specified datafile
        Copy other attributes like .vtk if there is any.

        Parameters
        ----------
        datafile : nextnano.outputs.DataFileTemplate object
        """

        self.metadata = datafile.metadata
        self.coords = datafile.coords
        self.variables = datafile.variables
        if hasattr(datafile, "vtk"):
            self.vtk = datafile.vtk

    def get_loader(self):
        pass

    def export(self, filename, format):
        raise NotImplementedError("Exporters are not implemented yet")


class DataFile(DataFileTemplate):
    def __init__(self, fullpath, product=None, **loader_kwargs):
        super().__init__(
            fullpath, product=product, **loader_kwargs
        )  # **loader_kwargs) #, FirstVarIsCoordFlag = False

    def get_loader(self):
        if self.product:
            loader = defaults.get_DataFile(self.product)
        else:
            print(
                "[Warning] nextnano product is not specified: nextnano++, nextnano3, nextnano.NEGF or nextnano.MSB"
            )
            print(
                "[Warning] Autosearching for the best loading method. Note: The result may not be correct"
            )
            loader = self.find_loader()
        return loader

    def find_loader(self):
        from nextnanopy.nnp.outputs import DataFile as DataFile_nnp
        from nextnanopy.nn3.outputs import DataFile as DataFile_nn3
        from nextnanopy.negf.outputs import DataFile as DataFile_negf
        from nextnanopy.msb.outputs import DataFile as DataFile_msb

        Dats = [DataFile_nn3, DataFile_nnp, DataFile_negf, DataFile_msb]
        for Dati in Dats:
            try:
                df = Dati(self.fullpath)
                if "" in df.variables.keys():
                    continue
                else:
                    break
            except:
                pass
        loader = Dati
        return loader

    def plot(self, legend=False, y_axis_name="", subplots=False):
        import matplotlib.pyplot as plt

        if len(self.coords) == 0:
            fig, ax = plt.subplots()
            if len(self.variables) > 1:
                plot_coord = self.variables[0]
                # plot_var = self.variables[1:]
                # for var in plot_var:
                for i in range(1, len(self.variables) - 1):
                    var = self.variables[i]
                    ax.plot(plot_coord.value, var.value, label=var.name)
                ax.set_xlabel(f"{plot_coord.name}[{plot_coord.unit}]")
            else:
                for var in self.variables:
                    ax.plot(var.value, label=var.name)
            ax.set_ylabel(f"{y_axis_name}[{var.unit}]")
        elif len(self.coords) == 1:
            fig, ax = plt.subplots()
            x_coord = self.coords[0]
            x_label_name = x_coord.name
            x_label_unit = x_coord.unit
            x_value = x_coord.value
            ax.set_xlabel(f"{x_label_name}[{x_label_unit}]")
            for var in self.variables:
                ax.plot(x_value, var.value, label=var.name)
            ax.set_ylabel(f"{y_axis_name}[{var.unit}]")
            if legend:
                plt.legend()
        elif len(self.coords) == 2:
            x = self.coords[0]
            y = self.coords[1]
            number_of_var = len(self.variables)
            if number_of_var < 2:
                fig, ax = plt.subplots()
                ax.set_xlabel(f"{x.name}[{x.unit}]")
                ax.set_ylabel(f"{y.name}[{y.unit}]")
                var = self.variables[0]
                im = ax.pcolormesh(
                    x.value,
                    y.value,
                    var.value,
                    shading="auto",
                    label=f"{var.name}[{var.unit}]",
                )
                ax.set_title(f"{var.name}[{var.unit}]")
                fig.colorbar(im, ax=ax)
            else:
                if subplots:
                    fig, ax = plt.subplots(len(self.variables))
                    for i, var in enumerate(self.variables):
                        im = ax[i].pcolormesh(
                            x.value,
                            y.value,
                            var.value,
                            shading="auto",
                            label=f"{var.name}[{var.unit}]",
                        )
                        ax[i].set_xlabel(f"{x.name}[{x.unit}]")
                        ax[i].set_ylabel(f"{y.name}[{y.unit}]")
                        ax[i].set_title(f"{var.name}[{var.unit}]")
                        fig.colorbar(im, ax=ax[i])
                else:
                    for i, var in enumerate(self.variables):
                        fig, ax = plt.subplots()
                        im = ax.pcolormesh(
                            x.value,
                            y.value,
                            var.value,
                            shading="auto",
                            label=f"{var.name}[{var.unit}]",
                        )
                        ax.set_xlabel(f"{x.name}[{x.unit}]")
                        ax.set_ylabel(f"{y.name}[{y.unit}]")
                        ax.set_title(f"{var.name}[{var.unit}]")
                        fig.colorbar(im, ax=ax)
        else:
            raise NotImplementedError(
                "Preview plot feature is implemented only for datafiles with 2 or less coordinates"
            )
        return fig, ax

    def save(self, filepath, format="dat"):
        """
        Save the data from the DataFile instance to a specified file in various formats.

        Parameters:
            filepath (str): The file path where the data will be saved.
            format (str, optional): The format in which the data should be saved. Default is 'dat'.
                Supported formats:
                - 'dat' (for 1D files): Data is saved in a plain text format (.dat) with whitespace-separated values.
                - 'VTKAscii' (for 2D/3D files): Data is saved in VTK ASCII format (.vtk) suitable for visualization tools.
                - 'AvsAscii_one_file' (for 2D/3D files): Data is saved in AVS/Express ASCII format (.fld) for AVS/Express software.

        Raises:
            NotImplementedError: If the provided 'format' is not supported for saving.

        Notes:
            - For the 'dat' format, the DataFile instance should be one-dimensional (ndim=1).
            - The 'VTKAscii' format requires 'nextnanopy.utils.formatting' module for creating the VTK header.
            - The 'AvsAscii_one_file' format requires the 'write_avsascii_one_file' function.

        Example:
            # Assuming `data_file` is an instance of the `DataFile` class
            data_file.save('data_file.dat', format='dat')  # Save data in .dat format
            data_file.save('data_file.vtk', format='VTKAscii')  # Save data in VTK ASCII format
            data_file.save('data_file.fld', format='AvsAscii_one_file')  # Save data in AVS/Express ASCII format
        """
        # TODO  and AvsBinary (.fld)
        accepted_format = ["dat", "VTKAscii", "AvsAscii_one_file"]
        if format not in accepted_format:
            raise NotImplementedError(f"{format} format is not supported for saving")
        if format == "dat":
            ndim = len(self.coords)
            if ndim > 1:
                raise ValueError(
                    f"The DataFile is {ndim}-dimensional. More than 1 dimension is not supported for .dat extension"
                )
            with open(filepath, "w") as f:
                # Write headers
                for header in self.metadata["headers"]:
                    f.write(header)

                # Write data
                combined_data = [coord.value for coord in self.coords] + [
                    variable.value for variable in self.variables
                ]

                data = np.column_stack(combined_data).transpose()
                np.savetxt(f, data.T)
        elif format == "VTKAscii":
            from nextnanopy.utils.formatting import create_vtk_header

            header = create_vtk_header(
                len(self.coords), [len(coord.value) for coord in self.coords]
            )
            with open(filepath, "w") as file:
                # Write the header
                file.write(header)

                # Write the coordinates
                file.write("<Coordinates>")
                for coord in self.coords:
                    file.write(
                        f'<DataArray type="Float64" Name="{coord.name.upper()}_COORDINATES" NumberOfComponents="1" format="ascii">\n'
                    )
                    np.savetxt(file, coord.value, fmt="%.6f", delimiter="\t")
                    file.write("</DataArray>\n")
                if len(self.coords) == 2:  # add Z coordinate
                    file.write(
                        """<DataArray type="Float64" Name ="Z_COORDINATES" NumberOfComponents ="1" format="ascii">
                                        0
                                    </DataArray>\n"""
                    )
                file.write("</Coordinates>\n")

                # Write the data
                file.write("<PointData>\n")
                for variable in self.variables:
                    if variable.unit:
                        name = variable.name + f"[{variable.unit}]"
                    else:
                        name = variable.name
                    file.write(
                        f'<DataArray type="Float64" Name="{name}" NumberOfComponents="1" format="ascii">\n'
                    )
                    flatten_array = variable.value.flatten("F")

                    np.savetxt(file, flatten_array, fmt="%.6f", delimiter="\t")
                    file.write("</DataArray>\n")

                # Write the footer
                footer = """</PointData>
                            </Piece>
                            </RectilinearGrid>
                            </VTKFile>"""
                file.write(footer)
        elif format == "AvsAscii_one_file":
            write_avsascii_one_file(
                coordinates=self.coords, variables=self.variables, filename=filepath
            )
        self.filepath = filepath


class AvsAscii(Output):
    def __init__(self, fullpath, **loader_kwargs):
        super().__init__(fullpath)
        self.load()

    @property
    def fld(self):
        filename = self.filename_only + ".fld"
        return os.path.join(self.folder, filename)

    def load(self):
        self.load_raw_metadata()
        self.load_metadata()
        self.load_variables()
        self.load_coords()

    def load_raw_metadata(self):
        possible_keys = [
            "ndim",
            "dim",
            "nspace",
            "veclen",
            "data",
            "field",
            "label",
            "variable",
            "coord",
        ]
        info = []
        try:
            with open(self.fld, "r") as f:
                for line in f:
                    line = line.replace("\n", "")
                    line = line.strip()
                    try:
                        float(line)
                        break
                    except:
                        if line == "":
                            continue
                        # if line[0] != '#':
                        #     info.append(line)
                        if start_with_choice(line, *possible_keys):
                            info.append(line)
        except UnicodeDecodeError:
            with open(self.fld, "rb") as f:
                for line in f:
                    try:
                        line = line.decode("ascii")
                    except:
                        break
                    line = line.replace("\n", "")
                    line = line.strip()
                    if line == "":
                        continue
                    # if line[0] != '#':
                    #     info.append(line)
                    if start_with_choice(line, *possible_keys):
                        info.append(line)
        return info

    def load_metadata(self):
        info = self.load_raw_metadata()
        key_int = ["ndim", "dim1", "dim2", "dim3", "nspace", "veclen"]
        key_str = ["data", "field"]
        metadata = {}
        metadata["labels"] = []
        metadata["units"] = []
        metadata["variables"] = []
        metadata["coords"] = []
        metadata["dims"] = []
        for line in info:
            key, value = line.split(maxsplit=1)
            if value[0] == "=":
                value = value.replace("=", "")
                value = value.strip()
                if key in key_int:
                    value = int(value)
                    metadata[key] = value
                elif key == "label":
                    labels = value.split()
                    for label in labels:
                        label, unit = best_str_to_name_unit(label, default_unit=None)
                        metadata["labels"].append(label)
                        metadata["units"].append(unit)
                else:
                    value = str(value)
                    metadata[key] = value

                if key[:3] == "dim":
                    metadata["dims"].append(metadata[key])

            else:
                if key == "variable":
                    vm = values_metadata(line)
                    vm["file"] = os.path.join(self.folder, vm["file"])
                    vm["size"] = np.prod(metadata["dims"])
                    metadata["variables"].append(vm)
                elif key == "coord":
                    vm = values_metadata(line)
                    vm["file"] = os.path.join(self.folder, vm["file"])
                    num = vm["num"]
                    vm["size"] = metadata[f"dim{num}"]
                    metadata["coords"].append(vm)

        self.metadata = metadata
        return metadata

    def load_variables(self):
        meta = self.metadata
        variables = DictList()
        for vmeta, label, unit in zip(meta["variables"], meta["labels"], meta["units"]):
            values = load_values(
                file=vmeta["file"],
                filetype=vmeta["filetype"],
                datatype=meta["data"],
                skip=vmeta["skip"],
                offset=vmeta["offset"],
                stride=vmeta["stride"],
                size=vmeta["size"],
            )
            values = reshape_values(values, *meta["dims"])
            var = Variable(name=label, value=values, unit=unit, metadata=vmeta)
            variables[var.name] = var
        self.variables = variables
        return variables

    def load_coords(self):
        meta = self.metadata
        coords = DictList()
        for vmeta in meta["coords"]:
            values = load_values(
                file=vmeta["file"],
                filetype=vmeta["filetype"],
                datatype=meta["data"],
                skip=vmeta["skip"],
                offset=vmeta["offset"],
                stride=vmeta["stride"],
                size=vmeta["size"],
            )
            ax = coord_axis(vmeta["num"])
            unit = "nm"
            var = Coord(
                name=ax, value=values, unit=unit, dim=vmeta["num"] - 1, metadata=vmeta
            )
            coords[var.name] = var
        self.coords = coords
        return coords


class Vtk(Output):
    def __init__(self, fullpath, **loader_kwargs):
        super().__init__(fullpath)
        self.load()

    def load(self):
        self.vtk = pv.read(self.fullpath)
        self.load_coords()
        self.load_variables()

    def load_coords(self):
        for i, coord in enumerate(["x", "y", "z"]):
            if not hasattr(self.vtk, coord):
                continue
            value = getattr(self.vtk, coord)
            if value.size == 1:
                continue
            self.coords[coord] = Coord(name=coord, value=value, unit=None, dim=i)

    def load_variables(self):
        for _name in self.vtk.array_names:
            name, unit = best_str_to_name_unit(_name, default_unit=None)
            value = (
                np.array(self.vtk[_name])
                .reshape(self.vtk.dimensions, order="F")
                .squeeze()
            )
            self.variables[name] = Variable(name=name, value=value, unit=unit)


class Dat(Output):
    def __init__(self, fullpath, **loader_kwargs):
        super().__init__(fullpath)
        self.load(**loader_kwargs)

    def load(self, **loader_kwargs):
        self.load_metadata(**loader_kwargs)
        self.load_data()

    def _get_headers(self):
        headers = []
        with open(self.fullpath, "r") as f:
            for line in f:
                try:
                    float(line.split()[0])
                    break
                except:
                    headers.append(line)
        return headers

    def _get_nb_columns(self):
        meta = self.metadata
        with open(self.fullpath, "r") as f:
            for i, line in enumerate(f):
                if i < meta["skip_rows"]:
                    continue
                line = line.replace("\n", "").strip().split()
                arr = np.array(line)
                break
        return arr.size

    def load_metadata(self, FirstVarIsCoordFlag=True):
        headers = self._get_headers()
        self.metadata["headers"] = headers
        self.metadata["skip_rows"] = len(headers)
        nb = self._get_nb_columns()
        self.metadata["nb_columns"] = nb

        if len(headers) == 0:
            raise NotImplementedError(".dat file without header")
        else:
            header = headers[-1]  # take the last one by default

        header = self._split_into_columns(header, nb)
        if header is None:
            raise RuntimeError(
                f"Can not load the datafile {self.fullpath}. The number of columns is not consistent"
            )

        columns_and_units = [self._extract_name_and_unit(column) for column in header]

        # diff = len(columns) - len(units)
        # if len(columns) > nb:
        #     columns = columns[0:nb]
        # elif diff > 0:
        #     empty = [""] * diff
        #     units.extend(empty)
        ndim = 0
        dkeys = []

        for i, (column, unit) in enumerate(columns_and_units):
            self.metadata[i] = {"name": column, "unit": unit}
            if FirstVarIsCoordFlag:
                ndim += 1
                dkeys.append(i)
                FirstVarIsCoordFlag = False
        self.metadata["ndim"] = ndim
        self.metadata["dkeys"] = dkeys
        return self.metadata

    def load_data(self):
        data = []
        meta = self.metadata
        # with open(self.fullpath, 'r') as f:
        #     for i, line in enumerate(f):
        #         if i < meta['skip_rows']:
        #             continue
        #         line = line.replace('\n', '').strip().split()
        #         if line:
        #             data.append(line)
        data = np.loadtxt(self.fullpath, skiprows=meta["skip_rows"])
        data = np.array(data, dtype=float).T  # columns 1st index
        coords, variables = DictList(), DictList()
        dims = []
        for i, values in enumerate(data):
            vm = meta[i]
            if i in meta["dkeys"]:
                # values = np.unique(values)
                dims.append(values.size)
                var = Coord(name=vm["name"], unit=vm["unit"], dim=i, value=values)
                coords[var.name] = var
            else:
                if dims:
                    values = values.reshape(*dims)
                var = Variable(name=vm["name"], unit=vm["unit"], value=values)
                variables[var.name] = var
        self.coords = coords
        self.variables = variables
        return coords, variables

    @staticmethod
    def _split_into_columns(header, expected_num_columns):
        # First try to split by any number of spaces
        columns = re.split(r"\s+", header.strip())

        # Check if the number of columns matches the expected number
        if len(columns) == expected_num_columns:
            return columns

        # If not matching, try splitting by two or more spaces or tabs
        columns = re.split(r"\s{2,}|\t+", header)

        # Final check to ensure we have the right number of columns (optional)
        if len(columns) != expected_num_columns:
            None
        return columns

    @staticmethod
    def _extract_name_and_unit(column):
        # Check if there is a unit in brackets
        match = re.search(r"\[(.*?)\]", column)
        if match:
            # Extract unit and remove it along with brackets from the column
            unit = match.group(1)
            name = column.replace(f"[{unit}]", "").strip()
        else:
            name, unit = column.strip(), ""
        return name, unit


def coord_axis(dim):
    dim = str(dim)
    axes = {"1": "x", "2": "y", "3": "z"}
    return axes[dim]


def values_metadata(line):
    """Return a dict for: kind, num, file, filetype, skip, offset, stride"""
    metadata = {}
    kind, num, rest = line.split(maxsplit=2)
    metadata["kind"] = kind
    metadata["num"] = int(num)
    raw_rest = rest.split("=")
    raw_rest = [r.strip().split() for r in raw_rest]
    rest = []
    for ri in raw_rest:
        rest.extend(ri)
    keys = rest[0::2]
    values = rest[1::2]
    for key, value in zip(keys, values):
        key = key.strip()
        value = value.strip()
        if key in ["num", "skip", "offset", "stride"]:
            value = int(value)
        metadata[key] = value
    return metadata


def load_values(
    file, filetype="ascii", datatype="double", skip=0, offset=0, stride=1, size=None
):
    """Return flat array of floating values"""
    if filetype == "ascii":

        stop = skip + size if size != None else None
        with open(file, "rb") as f:
            lines = islice(f, skip, stop, 1)
            values = [line.decode("ascii").replace("\n", "").strip().split()[offset] for line in lines]
    elif filetype == "binary":
        datatypes = {"double": "d"}
        datatype_sizes = {"double": 8}
        datatype_size = datatype_sizes[datatype]
        with open(file, "rb") as datafile:
            datafile.seek(skip)
            data = datafile.read(size * datatype_size)
            values = []
            iteration = struct.iter_unpack(datatypes[datatype], data)
            for i in iteration:
                values.append(i[0])

    else:
        raise ValueError("filetype is not recognized or implemented")
    return np.array(values, dtype=float)


def reshape_values(values, *dims):
    dims = np.flip(dims)
    shape = tuple([dim for dim in dims])
    values = np.reshape(values, shape)
    return np.transpose(values)


def write_avsascii_one_file(coordinates, variables, filename, binary=False):
    if binary:
        filetype = "binary"
        numpy_encoding = "bytes"
    else:
        filetype = "ascii"
        numpy_encoding = "bytes"
    # Validate input
    if len(coordinates) < 2 or len(coordinates) > 3:
        raise ValueError("Coordinates array must have exactly 2 or 3 objects.")

    num_coords = len(coordinates)
    num_variables = len(variables)

    # Calculate skip values
    # TODO calculate skip values (in bytes?) for binary output
    header_up_to_first_label = 9
    header_lines = (
        header_up_to_first_label + num_variables + 1 + num_variables + num_coords + 1
    )  # each 1 is skipped line
    coord_skip_values = [header_lines]
    for i in range(num_coords - 1):
        coord_skip_values.append(
            coord_skip_values[-1] + coordinates[i].value.size + 1
        )  # 1 for empty line after each coord

    data_skip = coord_skip_values[-1] + coordinates[-1].value.size + 1

    # Calculate the number of lines each variable takes
    number_of_lines_for_var = np.prod([coord.value.size for coord in coordinates])

    # Open file for writing
    with open(filename, "w") as file:
        # Write header
        file.write("# AVS/Express field file\n")
        file.write("#\n")
        file.write(f"ndim = {num_coords}\n")

        for i in range(num_coords):
            file.write(f"dim{i+1} = {coordinates[i].value.size}\n")

        file.write("nspace = {}\n".format(num_coords))
        file.write(f"veclen = {num_variables}\n")
        file.write("data = double\n")
        file.write("field = rectilinear\n")

        # Write variable labels and units
        for i in range(num_variables):
            file.write(f"label = {variables[i].name}[{variables[i].unit}]\n")

        # Write an empty line after the last label
        file.write("\n")

        # Write variable file paths
        for i, var in enumerate(variables):
            file.write(
                f"variable {i + 1} file={filename} filetype={filetype} skip={data_skip + i * number_of_lines_for_var} offset=0 stride=1\n"
            )

        # Write coordinate file paths
        for coord, skip_value, i in zip(
            coordinates, coord_skip_values, range(num_coords)
        ):
            file.write(
                f"coord {i+1} file={filename} filetype={filetype} skip={skip_value} offset=0 stride=1\n"
            )

        # Write an empty line after the last coord line
        file.write("\n")

        # Write data for coordinates
        for coord in coordinates:
            np.savetxt(
                file, coord.value.flatten("F"), fmt="%.8f", encoding=numpy_encoding
            )
            file.write("\n")

        # Write data for variables
        for var in variables:
            np.savetxt(
                file, var.value.flatten("F"), fmt="%.8f", encoding=numpy_encoding
            )



import sys, os
import subprocess
from pathlib import Path
import queue
import threading
import warnings
from nextnanopy.utils.misc import get_filename, mkdir_if_not_exist
from nextnanopy import defaults

from nextnanopy.utils.formatting import generate_command


def command(
    inputfile,
    exe,
    license,
    database,
    outputdirectory,
    **opt_kwargs,
):
    kwargs = dict(
        inputfile=inputfile,
        exe=exe,
        license=license,
        database=database,
        outputdirectory=outputdirectory,
    )
    kwargs.update(opt_kwargs)
    product = defaults.input_file_type(inputfile)
    cmd = defaults.get_command(product)

    # warn if output path might be too long
    outdir_len = len(str(outputdirectory))
    tooLongPath = (product in ["nextnano3", "nextnano++"]) and outdir_len + 80 > 260  # TODO: how long is the minimal path appended by nn3/nnp simulations?
    tooLongPathNEGF = (product in ["nextnano.NEGF", "nextnano.NEGF_classic"]) and outdir_len + 80 > 260
    if tooLongPath or tooLongPathNEGF:
        warnings.warn(
            "The output path might be too long on Windows 10 (maximum 260 characters). Consider abbreviating your input file name and/or sweep variables..."
        )
    return cmd(**kwargs)


def send(cmd, cwd=os.getcwd()):
    PIPE = subprocess.PIPE
    return subprocess.Popen(
        cmd, stdout=PIPE, stderr=PIPE, close_fds=True, shell=True, cwd=cwd
    )


def read_output(pipe, funcs):
    for line in iter(pipe.readline, b""):
        for func in funcs:
            func(line)
    pipe.close()


def write_output(get, filepath, show=True):
    with open(filepath, "w", newline="") as f:
        for line in iter(get, None):
            line = str(line, "utf-8", errors="ignore")
            if show:
                sys.stdout.write(line)
            f.write(line)


def start_log(process, filepath, show=True, parallel=False):
    q = queue.Queue()
    out, err = [], []
    tout = threading.Thread(
        target=read_output, args=(process.stdout, [q.put, out.append])
    )
    terr = threading.Thread(
        target=read_output, args=(process.stderr, [q.put, err.append])
    )
    twrite = threading.Thread(target=write_output, args=(q.get, filepath, show))
    for t in (tout, terr, twrite):
        t.daemon = False
        t.start()

    if parallel:
        return q, tout, terr
    else:
        process.wait()
        for t in (tout, terr):
            t.join()
        q.put(None)
        return q, tout, terr


def execute(
    inputfile,
    exe,
    license,
    database,
    outputdirectory,
    show_log=True,
    parallel=False,
    **kwargs,
):
    filename = get_filename(inputfile, ext=False)
    inputfile = Path(inputfile).resolve()
    outputdirectory = Path(outputdirectory) / filename
    mkdir_if_not_exist(outputdirectory)
    logfile = outputdirectory / f"{filename}.log"
    cmd = command(inputfile, exe, license, database, outputdirectory, **kwargs)
    cwd = os.getcwd()
    exe = Path(exe)
    wdir, executable = exe.parent, exe.name  # nn3 assumes wdir at one folder upper than the executable

    # validate configuration of executable path
    if executable == "":
        raise FileNotFoundError(f"Executable path is empty! Check nextnanopy.config")

    if (not exe.is_file()) or (not wdir.is_dir()):
        raise FileNotFoundError(
            f"Executable path is invalid: {exe}\nCheck nextnanopy.config"
        )
    process = send(cmd, cwd=wdir)
    queue, tout, terr = start_log(process, logfile, show_log, parallel=parallel)
    os.chdir(cwd)
    info = {
        "process": process,
        "outputdirectory": outputdirectory,
        "filename": filename,
        "logfile": logfile,
        "cmd": cmd,
        "wdir": wdir,
        "queue": queue,
        "tout": tout,
        "terr": terr,
    }
    return info


def run_script(script, kwargs=None, show_log=True):
    """
    The function runs a python script with given arguments. Output is stored in the file script_name.log
    Parameters
    ----------
    script: str
        path to the python script
    kwargs: dict
        optional parameters, {keyword:argument,keyword:argument}
        for a keyword without argument leave argument as an empty string ''

        EXAMPLE: {'-o': 'my_output_folder','-p':''} will be converted to '-o my_output_folder -p'
    show_log - bool
        show the log in console output, default is True

    Returns
    -------
    process: subprocess.POPEN
    """
    args = [[sys.executable, script]]
    if kwargs:
        for key in kwargs.keys():
            args.append([key, kwargs[key]])
    cmd = generate_command(args)
    process = send(cmd)
    logfile = Path.cwd() / f"{Path(script).name}.log"
    start_log(process, logfile, show_log)
    return process



from nextnanopy.utils.mycollections import DictList
from nextnanopy.inputs import InputFileTemplate
from nextnanopy.nn3.defaults import parse_nn3_variable, is_nn3_input_text, is_nn3_variable, InputVariable_nn3


class InputFile(InputFileTemplate):
    def load_variables(self):
        variables = DictList()
        for i, line in enumerate(self.raw_lines):
            if not is_nn3_variable(line):
                continue
            name, value, comment = parse_nn3_variable(line)
            var = InputVariable_nn3(name=name, value=value, comment=comment, metadata={'line_idx': i})
            variables[var.name] = var
        self.variables = variables
        return self.variables

    def validate(self):
        if not is_nn3_input_text(self.raw_text):
            raise ValueError(f'Not valid nextnano3 input file')



import unittest
from nextnanopy.utils.mycollections import DictList


class TestDictList(unittest.TestCase):

    def test_indexes(self):
        dl = DictList(a=3, b='test')
        self.assertEqual(list(dl.keys())[0], 'a')
        self.assertEqual(list(dl.keys())[1], 'b')
        self.assertEqual(list(dl.values())[0], 3)
        self.assertEqual(list(dl.values())[1], 'test')
        self.assertEqual(dl[0], dl['a'])
        self.assertEqual(dl[0], 3)
        self.assertEqual(dl[1], dl['b'])
        self.assertEqual(dl[1], 'test')

    def test_loop(self):
        dl = DictList(a=3, b='test')
        for value, expected in zip(dl, dl.values()):
            self.assertEqual(value, expected)


if __name__ == '__main__':
    unittest.main()



import numpy as np
from nextnanopy.utils.mycollections import DictList
from nextnanopy.outputs import Output, AvsAscii, Vtk, DataFileTemplate, Dat
from nextnanopy.utils.datasets import Variable, Coord
import re
import os


class DataFile(DataFileTemplate):
    def __init__(self, fullpath, **loader_kwargs):
        super().__init__(fullpath, product="nextnano.NEGF")
        self.load(**loader_kwargs)

    def get_loader(self):
        if self.extension in [".v", ".fld", ".coord"]:
            loader = AvsAscii
        elif self.extension == ".vtr":
            loader = Vtk
        elif self.extension == ".txt":
            raise NotImplementedError(
                f"Loading nextnano.NEGF datafiles with extension.txt is not implemented yet"
            )
        elif self.extension == ".dat":
            loader = Dat
        else:
            raise NotImplementedError(
                f"Loading datafile with extension {self.extension} is not implemented yet"
            )
        return loader


def get_iv(path=""):
    piv = os.path.join(path, "Current_vs_Voltage.dat")
    return np.loadtxt(piv, skiprows=1, delimiter="\t", unpack=True)


def get_WannierStark_on(folder):
    ws = np.loadtxt(
        folder + r"\WannierStark\\WannierStark_statesOn.dat",
        skiprows=2,
        delimiter="\t",
        unpack=True,
    )
    #  print('Number of states: ', len(ws)-2)
    return ws


def get_WannierStark(folder):
    ws = np.loadtxt(
        folder + r"\WannierStark\\WannierStark_states.dat",
        skiprows=2,
        delimiter="\t",
        unpack=True,
    )
    #  print('Number of states: ', len(ws)-2)
    return ws


def get_WannierStark_norm(folder, scaling_factor=1):
    ws = np.loadtxt(
        folder + r"\WannierStark\\WannierStark_states.dat",
        skiprows=2,
        delimiter="\t",
        unpack=True,
    )
    #  print('Number of states: ', len(ws)-2)
    norm = min(ws[1])
    z = ws[0]
    pot = ws[1] - norm
    ws_norm = ws[2:] - norm
    ws_norm_scal = scale_wf(ws_norm, scaling_factor)
    return z, pot, ws_norm_scal


def get_WannierStark_norm_cpp(folder, scaling_factor=1):
    ws = np.loadtxt(
        folder + r"\EnergyEigenstates\\EigenStates.dat",
        skiprows=2,
        delimiter="\t",
        unpack=True,
    )
    #  print('Number of states: ', len(ws)-2)
    norm = min(ws[1])
    z = ws[0]
    pot = ws[1] - norm
    ws_norm = ws[2:] - norm
    ws_norm_scal = scale_wf(ws_norm, scaling_factor)
    return z, pot, ws_norm_scal


def scale_wf(wf_input, factor):
    scaled = np.copy(wf_input)
    for i, cur in enumerate(wf_input):
        mi = min(cur)
        ma = max(cur)
        scaled[i] = np.interp(cur, [mi, ma], [mi, factor * (ma - mi) + mi])

    return scaled



from nextnanopy.utils.mycollections import DictList
from nextnanopy.inputs import InputFileTemplate
from nextnanopy.negf.defaults import (
    is_negf_classic_input_text,
    InputVariable_NEGF,
    parse_negf_variable_name,
)
from nextnanopy.utils.formatting import autofmt_variable_value
import xml.etree.ElementTree as ET


class InputFile(InputFileTemplate):
    def load_variables(self):
        # parser = ET.XMLParser(target=ET.TreeBuilder(insert_comments=True)) #Does not supported in python3.7
        parser = ET.XMLParser()
        root = ET.fromstring(self.text, parser=parser)
        varsection = root.find("Variables")
        variables = DictList()
        if not varsection:
            self.variables = variables
            return self.variables
        vars = varsection.findall("Constant")

        for var in vars:
            name_element = var.find("Name")
            name = parse_negf_variable_name(name_element.text)
            comment = name_element.get("Comment")
            value_element = var.find("Value")
            value = autofmt_variable_value(value_element.text)
            unit = value_element.get("Unit")
            var = InputVariable_NEGF(name=name, value=value, unit=unit, comment=comment)
            variables[name] = var
        self.variables = variables
        return self.variables

    @property
    def lines(self):
        root = ET.fromstring(self.raw_text)
        varsection = root.find("Variables")
        variables = self.variables
        if varsection:
            vars = varsection.findall("Constant")
            for var_element in vars:
                for variable in variables:
                    name_element = var_element.find("Name")
                    if name_element.text == "$" + variable.name:
                        value_element = var_element.find("Value")
                        value_element.text = str(variable.value)
                        name_element.set("Comment", variable.comment)
                        value_element.set("Unit", variable.unit)

        xmlstr = ET.tostring(root, encoding="unicode", method="xml")
        lines_result = xmlstr.splitlines()
        return lines_result

    def validate(self):
        if not is_negf_classic_input_text(self.raw_text):
            raise ValueError(f"Not valid nextnano.NEGF input file")



from nextnanopy.utils.datasets import InputVariable
from nextnanopy.utils.formatting import (
    str_to_path,
    _path,
    pattern_in_file,
    is_variable,
    parse_variable,
    generate_command,
    pattern_in_text,
)
from collections import OrderedDict


fmt_classic = {
    "var_char": "$",
    "com_char": "<!--",
    "input_pattern": "<Simulation",
}

fmt = {
    "var_char": "$",
    "com_char": "#",
    "input_pattern": "nextnano.NEGF{",
}

config_validator = {
    "exe": str_to_path,
    "license": str_to_path,
    "database": str_to_path,
    "outputdirectory": str_to_path,
    "threads": int,
}

config_default = {
    "exe": "",
    "license": "",
    "database": "",
    "outputdirectory": "",
    "threads": 0,
}


def command_negf_classic(
    inputfile,
    exe,
    license,
    database,
    outputdirectory,
    threads=0,
    **kwargs,
):
    kwargs = OrderedDict(
        exe=[_path(exe), ""],
        inputfile=[_path(inputfile), ""],
        outputdirectory=[_path(outputdirectory), ""],
        database=[_path(database), ""],
        license=[_path(license), ""],
        threads=["-threads", threads],
    )
    return generate_command(kwargs.values())


def command_negf(
    inputfile,
    exe,
    license,
    database,
    outputdirectory,
    threads=0,
    debug_ouptut_specifications=None,
    **kwargs,
):
    kwargs = OrderedDict(
        exe=[_path(exe), ""],
        inputfile=["--input-file", _path(inputfile)],
        outputdirectory=["--outputdirectory", _path(outputdirectory)],
        database=["--database", _path(database)],
        license=["--license", _path(license)],
        threads=["--threads", threads],
    )
    if debug_ouptut_specifications is not None:
        kwargs["debug"] = ["--debug", debug_ouptut_specifications]
        kwargs.move_to_end("debug")
    return generate_command(kwargs.values())


def is_negf_variable(text):
    return is_variable(text, var_char=fmt_classic["var_char"])


# def parse_negf_variable(text):
#    return parse_variable(text, var_char=fmt['var_char'], com_char=fmt['com_char'])


def parse_negf_variable_name(text):
    if text.startswith(fmt_classic["var_char"]):
        return text[1:]
    else:
        raise ValueError(
            f"variable name in NEGF should start with {fmt_classic['var_char']}"
        )


def is_negf_classic_input_file(fullpath):
    return pattern_in_file(fullpath, fmt_classic["input_pattern"])


def is_negf_input_file(fullpath):
    return pattern_in_file(fullpath, fmt["input_pattern"])


def is_negf_classic_input_text(text):
    return pattern_in_text(text, fmt_classic["input_pattern"])


def is_negf_input_text(text):
    return pattern_in_text(text, fmt["input_pattern"])


class InputVariable_NEGF(InputVariable):
    var_char = fmt_classic["var_char"]

    @property
    def text(self):
        t = f"{self.var_char}{self.name} = {self.value} {self.unit}"
        if self.comment:
            t = f"{t} Comment: {self.comment}"
        return t



import setuptools
from packaging.version import Version, InvalidVersion
from importlib import import_module


VERSION = "1.0.4"

with open("README.md", "r") as fh:
    long_description = fh.read()

extras = {
    "gdspy": ("gdspy", "1.6", "pip"),
    "matplotlib": ("matplotlib", "3.2", "conda"),
    "cycler": ("cycler", "0.10", "conda"),
    "shapely": ("shapely", "1.7", "conda"),
}
extras_require = {k: ">=".join(v[0:2]) for k, v in extras.items()}

install_requires = [
    "numpy>=1.18",
    "pyvista>=0.27",
]

setuptools.setup(
    name="nextnanopy",
    version=VERSION,
    author="nextnano GmbH",
    author_email="python@nextnano.com",
    license="BSD-3-Clause",
    description="Useful tools to interface the nextnano software (https://www.nextnano.com/)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="nextnano",
    url="https://github.com/nextnanopy/nextnanopy",
    packages=setuptools.find_packages(exclude=["tests"]),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Intended Audience :: Customer Service",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
    ],
    python_requires=">=3.10",
    install_requires=install_requires,
    extras_require=extras_require,
    package_data={"nextnanopy.utils.styles": ["*.mplstyle"]},
)

# Code below adapted from QCoDeS (https://qcodes.github.io/)

version_template = """
*****
***** package {0} must be at least version {1}.
***** Please upgrade it (pip install -U {0} or conda install {0})
***** in order to use {2}
***** Recommended method: {3}
*****
"""

missing_template = """
*****
***** package {0} not found
***** Please install it (pip install {0} or conda install {0})
***** Recommended: {2} install {0}
***** in order to use {1}
*****
"""

valueerror_template = """
*****
***** package {0} version not understood
***** Please make sure the installed version ({1})
***** is compatible with the minimum required version ({2})
***** in order to use {3}
*****
"""

othererror_template = """
*****
***** could not import package {0}. Please try importing it from
***** the commandline to diagnose the issue.
*****
"""

# now test the versions of extras
for extra, (module_name, min_version, install_method) in extras.items():
    try:
        module = import_module(module_name.lower())
        try:
            if Version(module.__version__) < Version(min_version):
                print(
                    version_template.format(
                        module_name, min_version, extra, install_method
                    )
                )
        except InvalidVersion:
            print(
                valueerror_template.format(
                    module_name, module.__version__, min_version, extra
                )
            )
    except ImportError:
        print(missing_template.format(module_name, extra, install_method))
    except Exception:
        print(othererror_template.format(module_name))



def is_variable(text, var_char):
    boolean = False
    text = str(text).strip()
    if not text:
        boolean = False
    elif text[0] == var_char and '=' in text:
        boolean = True
    return boolean


def parse_variable(text, var_char, com_char):
    text = str(text).strip()
    if text[0] == var_char:
        text = str(text[1:])
    name, right = text.split('=', maxsplit=1)
    if com_char in right:
        value, comment = right.split(com_char, maxsplit=1)
    else:
        value = right
        comment = ''
    name, value, comment = [_str.strip() for _str in [name, value, comment]]
    value = autofmt_variable_value(value)
    return name, value, comment


def autofmt_variable_value(value):
    cases = [int, float, str]
    for case in cases:
        try:
            value = case(value)
            break
        except:
            pass
    return value


def pattern_in_file(fullpath, input_pattern):
    with open(fullpath, 'r') as f:
        for line in f:
            if input_pattern in line:
                return True
    return False


def pattern_in_text(text, input_pattern):
    if input_pattern in text:
        return True
    return False


def generate_command(args):
    cmd = []
    for case in args:
        arg, value = case
        _a, _v = _bool(arg), _bool(value)
        if not _a:
            continue
        elif _a and not _v:
            cmdi = f"{arg}"
        else:
            cmdi = f"{arg} {value}"
        cmd.append(cmdi)
    cmd = ' '.join(cmd)
    return cmd


def text(init='', mid='', end='', fmt=lambda i, m, f: i + m + f):
    return fmt(str(init), str(mid), str(end))


def text_to_lines(text):
    for line in str(text).split('\n'):
        yield line


def lines_to_text(*lines):
    ls = '\n'.join([str(t) for t in lines])
    return ls


def modify_lines_in_text(init='', mid='', end=''):
    ls = text_to_lines(mid)
    new = [text(init, line, end) for line in ls]
    return lines_to_text(*new)


def paragraph(content):
    return text('\n', content, '\n')


def preview(text, nums=True):
    lines = text_to_lines(text)
    for i, line in enumerate(lines):
        if nums:
            print(f'{i} {line}')
        else:
            print(f'{line}')


def str_to_bool(string,
                true_cases=['true', 't', 'yes', 'y', '1'],
                false_cases=['false', 'f', 'no', 'n', '0'],
                ):
    str_low = string.lower()
    if str_low in true_cases:
        value = True
    elif str_low in false_cases:
        value = False
    else:
        raise ValueError('Ambiguos string to be converted to boolean')
    return value


def str_to_path(string):
    path = r'{}'.format(string)
    return path


def _path(path):
    if path:
        path = r'{}'.format(path)
        path = f'"{path}"'
    return path


def _bool(_in):
    if _in == 0:
        return True
    else:
        return bool(_in)


def split_by_pattern(_str, init, end):
    _str = str(_str)
    _l = []

    if init not in _str and end not in _str:
        return _l

    find_init = True
    str_old = str(_str)
    while True:
        if init not in str_old or end not in str_old:
            break
        if find_init:
            if init in str_old:
                find_init = False
        else:
            if end in str_old:
                idx_end = str_old.find(end) + 1
                _l.append(str_old[0:idx_end])
                str_old = str_old[idx_end:]
                find_init = True
    return _l


def str_to_name_unit(_str, init='[', end=']', default_unit=None, add_rest_to_name=False):
    if init in _str and end in _str:
        splitted = _str.split(init)
        name = splitted[0]
        rest = splitted[1]
        tail = rest.split(end)
        unit, rest = tail[0], tail[1:]
        name = name.strip()
        unit = unit.strip()
        if add_rest_to_name:
            name += ''.join(rest)
            name = name.strip()
    else:
        name = str(_str).strip()
        unit = default_unit
    return name, unit


def str_to_name_unit_with_rest(_str, init, end, default_unit=None):
    """
    For the weird definition like 'psi[eV]_real' and 'psi[eV]_imag'
    """
    name, unit = str_to_name_unit(_str, init, end, default_unit, add_rest_to_name=True)
    return name, unit


def best_str_to_name_unit(_str, default_unit=None):
    _str_copy = str(_str)
    unit_patterns = [
        ['[', ']'],
        ['(', ')'],
    ]
    name = str(_str).strip()
    unit = default_unit
    for pattern in unit_patterns:
        init, end = pattern
        if init in _str and end in _str:
            name, unit = str_to_name_unit_with_rest(_str, init=init, end=end, default_unit=None)
            break
    return name, unit

def create_vtk_header(dim : int, coord_dimensions : list):
    extent_str = ''
    for i in range(dim):
        extent_str+= f"1 {coord_dimensions[i]} "
    if dim==2:
        extent_str+= "1 1 "
    extent_str = extent_str[:-1]

    header = rf"""<VTKFile type="RectilinearGrid" version="0.1" format="ascii">
<RectilinearGrid WholeExtent="{extent_str}">
<Piece Extent="{extent_str}">
    """

    return header








from nextnanopy.utils.mycollections import DictList
from nextnanopy.inputs import InputFileTemplate
from nextnanopy.nnp.defaults import parse_nnp_variable, is_nnp_input_text, is_nnp_variable, InputVariable_nnp
from nextnanopy.utils.misc import savetxt
import os

class InputFile(InputFileTemplate):
    def load_variables(self):
        variables = DictList()
        for i, line in enumerate(self.raw_lines):
            if not is_nnp_variable(line):
                continue
            name, value, comment = parse_nnp_variable(line)
            var = InputVariable_nnp(name=name, value=value, comment=comment, metadata={'line_idx': i})
            variables[var.name] = var
        self.variables = variables
        return self.variables

    def load_content(self):
        parser = Parser()
        #parser.parse(self.raw_lines, mode='lines')
        #print(self.raw_lines)
        parser.parse(self.lines, mode = 'lines')
        # result = parser.result
        # return result
        self.content = parser.result
    # @property
    # def content(self):
    #     return self.load_content()

    def save(self, fullpath=None, overwrite=False, automkdir=True, content = False, temp=False):
        """
        content=True invokes the demo feature of saving self.content instead of self.test

        be aware that the content=True ignores comments
        """
        if fullpath is None:
            if temp:
                folder = self._get_temp_dir()
                fullpath = os.path.join(folder, self.filename)
            elif self.fullpath is None:
                raise ValueError('Please, specify a fullpath')
            else:
                fullpath = self.fullpath
        if content:
            text = self.content.__str__()
        else:
            text = self.text
        self.fullpath = savetxt(fullpath=fullpath, text=text, overwrite=overwrite, automkdir=automkdir)
        return self.fullpath



    def validate(self):
        if not is_nnp_input_text(self.raw_text):
            raise ValueError(f'Not a valid nextnano++ input file')



######### objects to create abstract object of

def content_to_lines(content_to_transform):
    if isinstance(content_to_transform, str):
        return content_to_transform.split(sep='\n')
    else:
        result = []
        for c in content_to_transform:
            result.extend(c.__str__().split(sep='\n'))
        return result


class Comment(object): #just to be able to check that entry is comment
    def __init__(self, text):
        self.text = text

    def __repr__(self):
        self.text.__repr__()

    def __str__(self):
        self.text.__str__()


class Entry(object):
    def __init__(self, content, intend_sign = '    '):
        self.content = content
        self.intend_sign = intend_sign

    @property
    def lines(self, intend_level = 0):
        if isinstance(self.content, str):
            text_lines = content_to_lines(self.content)
        else:
            text_lines = []
            for c in self.content:
                text_lines.extend(content_to_lines(c.__str__()))
        for i, _ in enumerate(text_lines):
            text_lines[i] = intend_level * self.intend_sign + text_lines[i]

        return text_lines



    def __repr__(self):
        result = ''
        for line in self.lines:
            result += line + '\n'
        return result

    @property
    def dict_representation(self):
        result = DictList()
        for i, entry in enumerate(self.content):
            if isinstance(entry, Block):
                result[entry.name] = entry
            else:
                result[f'_entry_{i}'] = entry
        return result

    def __getitem__(self, item):
        return self.dict_representation[item]

    def __setitem__(self, item, val):

        self.dict_representation[item] = val
        if isinstance(item, int):
            self.content[item] = val
        else:
            i = self.dict_representation.get_indx(item)
            self.content[i] = val

    def __delitem__(self, item):
        del self.dict_representation[item]
        if isinstance(item, int):
            del self.content[item]
        else:
            i = self.dict_representation.get_indx(item)
            del self.content[i]





class Block(Entry):
    def __init__(self, name, content, parent_block = None, intend_sign = '    '):
        self.content = content
        self.intend_sign = intend_sign
        self.name = name
        self.parent_block = parent_block

    @property
    def lines(self, intend_level=1):
        if isinstance(self.content, str):
            text_lines = content_to_lines(self.content)
        else:
            text_lines = []
            for c in self.content:
                text_lines.extend(content_to_lines(c.__str__()))
        for i, _ in enumerate(text_lines):
            text_lines[i] = intend_level * self.intend_sign + text_lines[i]

        text_lines.insert(0, self.name + '{')
        if self.content:
            text_lines.append('}')
        else:
            text_lines[-1]+='}'

        return text_lines



class Parser(object):
    def __init__(self):
        self.lines = []
        self.result = Entry([])

    def parse(self, input, mode = 'str'):
        if mode == 'str':
            self.lines = input.split(sep = '\n')
        elif mode == 'lines':
            self.lines = input
        self.delete_comments()
        self.replace_brackets()
        last_word = ''
        current_block = self.result
        level = 0
        for line in self.lines:
            words_in_line = line.split()
            content_line = ''
            for word in words_in_line:
                if word == '{':
                    level+=1
                    name_of_the_block = last_word
                    new_block = Block(name_of_the_block, [])
                    if content_line:
                        content_line_without_last_word = content_line.rsplit(' ', 1) #split to beginning and last word

                        if not content_line_without_last_word[1]:
                            content_line = '' #if only one word was in content_line - it is name of the block, so delete it from content line
                        else:
                            content_line = content_line_without_last_word[0]
                            current_block.content.append(content_line)
                    current_block.content.append(new_block)
                    new_block.parent_block  = current_block
                    current_block = new_block
                    content_line = ''

                elif word == '}':
                    if level<1:
                        raise ValueError('Incorrect string to parse. } closed before opening')
                    level = level -1
                    if content_line:
                        current_block.content.append(content_line)
                    current_block = current_block.parent_block
                    last_word = ''

                else:
                    content_line += word+' '
                    last_word = word
            if content_line:
                current_block.content.append(content_line)
        if level != 0:
            raise ValueError('Incorrect string to parse. { not closed')



    def clear(self):
        self.lines = []
        self.result = Entry([])




    def delete_comments(self):
        for i, _ in enumerate(self.lines):
            if self.lines[i].startswith('#'):
                self.lines[i] = ''
            else:
                self.lines[i] = self.lines[i].split(sep = '#')[0]

    def replace_brackets(self):
        for i, _ in enumerate(self.lines):
            self.lines[i] = self.lines[i].replace('{', ' { ')
            self.lines[i] = self.lines[i].replace('}', ' } ')
            self.lines[i] = self.lines[i].replace('[', ' [ ')
            self.lines[i] = self.lines[i].replace(']', ' ] ')
            self.lines[i] = self.lines[i].replace('(', ' ( ')


    def text(self):
        result = ''
        for line in self.lines:
            result += line + '\n'
        return result


"""Defines colors and styles for nextnano plots."""
from matplotlib.colors import LinearSegmentedColormap

NXT_BLUE = (19/256, 173/256, 181/256) #13adb5 
WILD_STRAWBERRY = (255/256, 41/256, 103/256) #ff2966
GREEN = (0/256, 155/256, 85/256) #009b55
DANDELION = (240/256, 200/256, 0) #f0c800

GOLDEN = (253/256, 188/256, 66/256) #fdbc42, not part of palette anymore, substituted by dandelion

_NXT_COLORS = [
    (1, 1, 1),  # white
    NXT_BLUE,
    (0, 0, 0),  # black
    WILD_STRAWBERRY,
    (1, 1, 1),  # white
]

_NXT_BLUE_COLORS = [
    (0, 0, 0),  # black
    NXT_BLUE,
    (1, 1, 1),  # white
]

_NXT_STRAWBERRY_COLORS = [
    (0, 0, 0),  # black
    WILD_STRAWBERRY,
    (1, 1, 1),  # white
]

NXT_COLORMAP = LinearSegmentedColormap.from_list("NXT", _NXT_COLORS, N=1042)
NXT_BLUE_COLORMAP = LinearSegmentedColormap.from_list("NXT_BLUE", _NXT_BLUE_COLORS, N=1042)
NXT_STRAWBERRY_COLORMAP = LinearSegmentedColormap.from_list("NXT_STRAWBERRY", _NXT_STRAWBERRY_COLORS, N=1042)

import matplotlib.pyplot as plt
import importlib.resources as pkg_resources

def use_nxt_style():
    with pkg_resources.path("nextnanopy.utils.styles", "nxt_style.mplstyle") as style_path:
        plt.style.use(style_path)


from nextnanopy.nnp.inputs import InputFile as InputFileTemplateNnp
from nextnanopy.msb.defaults import is_msb_input_text


class InputFile(InputFileTemplateNnp): # inherits from NNP InputFile
    def validate(self):
        if not is_msb_input_text(self.raw_text):
            raise ValueError(f'Not valid nextnano.MSB input file')



from nextnanopy.utils.datasets import InputVariable
from nextnanopy.utils.formatting import str_to_path, _path, pattern_in_file, is_variable, parse_variable, \
    generate_command, pattern_in_text
from collections import OrderedDict

fmt = {
    'var_char': '%',
    'com_char': '!',
    'input_pattern': '$end_simulation-dimension',
}

config_validator = {
    'exe': str_to_path,
    'license': str_to_path,
    'database': str_to_path,
    'threads': int,
    'outputdirectory': str_to_path,
    'debuglevel': int,
    'cancel': int,
    'softkill': int,
}

config_default = {
    'exe': '',
    'license': '',
    'database': '',
    'threads': 0,
    'outputdirectory': '',
    'debuglevel': 0,
    'cancel': -1,
    'softkill': -1,
}


"""
The option 'system' should be implemented once we convince that everyone uses the later version of nn3 than 2021_12_24.
Also the default value of debuglevel should be a negative integer. (i.e. ignore the flag)
We only need to comment out the 3 corresponding lines.

About the difference of '-...' and '--...', '--license' will also work and should be the new default.
For backwards compatibility, we keep '-license' for the moment.
"""
def command_nn3(
        inputfile,
        exe,
        license,
        database,
        outputdirectory,
        threads=0,
        debuglevel=0,
        #dubuglevel=-1,             # pending change
        cancel=-1,
        softkill=-1,
        #system='default',          # pending change
        **kwargs,
):
    kwargs = OrderedDict(
        exe=[_path(exe), ''],
        license=['-license', _path(license)],
        inputfile=['-inputfile', _path(inputfile)],
        database=['-database', _path(database)],
        threads=['-threads', threads],
        outputdirectory=['-outputdirectory', _path(outputdirectory)],
        debuglevel=['-debuglevel', debuglevel],
        cancel=['-cancel', cancel],
        softkill=['-softkill', softkill],
        #system=['-system', system],        # pending change
        no_file_options=[kwargs['no_file_options'], ''] if 'no_file_options' in kwargs else ['', ''],
    )
    return generate_command(kwargs.values())


def is_nn3_variable(text):
    return is_variable(text, var_char=fmt['var_char'])


def parse_nn3_variable(text):
    return parse_variable(text, var_char=fmt['var_char'], com_char=fmt['com_char'])


def is_nn3_input_file(fullpath):
    return pattern_in_file(fullpath, fmt['input_pattern'])


def is_nn3_input_text(text):
    return pattern_in_text(text, fmt['input_pattern'])


class InputVariable_nn3(InputVariable):
    var_char = fmt['var_char']
    com_char = fmt['com_char']



import numpy as np
from nextnanopy.utils.mycollections import DictList
from nextnanopy.outputs import Output, AvsAscii, Vtk, DataFileTemplate, Dat
from nextnanopy.nn3.defaults import (
    parse_nn3_variable,
    is_nn3_variable,
    InputVariable_nn3,
)
from nextnanopy.utils.datasets import Variable, Coord
from nextnanopy.utils.formatting import best_str_to_name_unit


class DataFile(DataFileTemplate):
    def __init__(self, fullpath, **loader_kwargs):
        super().__init__(fullpath, product="nextnano3")
        self.load(**loader_kwargs)

    def get_loader(self):
        if self.extension in [".v", ".fld", ".coord"]:
            loader = AvsAscii
        elif self.extension == ".vtr":
            loader = Vtk
        elif self.extension == ".txt":
            loader = self._find_txt_loader()
        elif self.extension == ".dat":
            loader = Dat
        else:
            raise NotImplementedError(
                f"Loading datafile with extension {self.extension} is not implemented yet"
            )
        return loader

    def _find_txt_loader(self):
        if self.filename_only in ["variables_input", "variables_database"]:
            loader = InputVariables
        elif self.filename_only == "materials":
            NotImplementedError(f"Loading materials.txt is not implemented yet")
        elif self.filename_only == "total_charges":
            NotImplementedError(f"Loading total_charges.txt is not implemented yet")
        else:
            raise NotImplementedError(f"Datafile {self.filename_only}.txt is not valid")
        return loader


class InputVariables(Output):
    def __init__(self, fullpath):
        super().__init__(fullpath)
        self.load()

    def load(self):
        self.load_raw()
        self.load_variables()

    def load_raw(self):
        with open(self.fullpath, "r") as f:
            self.raw_lines = f.readlines()

    def load_variables(self):
        variables = DictList()
        for i, line in enumerate(self.raw_lines):
            if not is_nn3_variable(line):
                continue
            name, value, comment = parse_nn3_variable(line)
            var = InputVariable_nn3(
                name=name, value=value, comment=comment, metadata={"line_idx": i}
            )
            variables[var.name] = var
        self.variables = variables
        return self.variables
